==> yass-1.07.tar.gz: yass-1.07/align.c <==
/*
 *  YASS 1.07
 *  Copyright (C) 2004,2005
 *  YASS Team
 *  Laurent Noe, Gregory Kucherov, Mikhail Roytberg, Christophe Valmir, Steven Corroy
 *  yass@loria.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software contains code derived from the GNU libavl library.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <stdio.h>
#include <stdlib.h>

#include "align.h"
#include "kword.h"
#include "tuple.h"
#include "prdyn.h"
#include "display.h"
#include "util.h"
#include "global_var.h"


#define FLAGSAVE() {                                \
 flagsave = (total_score >= MAX(0,mem_total_score));\
                                                    \
}

#define FLAGDROP() {                                        \
  flagdrop = (total_score  + gp_xdrop <  mem_total_score ); \
                                                            \
                                                            \
}


#define IF_FLAGDROP_MEMORISE_TMP_IN_DISPLAY_AND_RESET(){    \
 FLAGDROP();                                                \
 if ( flagdrop ){                                           \
    MEMORISE_TMP_IN_DISPLAY_AND_RESET();                    \
 }                                                          \
}



#define IF_FLAGSAVE_MEMORISE_IN_TMP()  \
{                                      \
  FLAGSAVE();                          \
  if ( flagsave ) {                    \
     MEMORISE_IN_TMP();                \
  }                                    \
}


#define MEMORISE_IN_TMP() {                              \
  mem_first_tuple_aligned = first_tuple_aligned;         \
  mem_last_tuple_aligned  = last_tuple_aligned;          \
  mem_total_score         = total_score;                 \
  /* mem_Mss = Mss;                                      \
   * mem_mss = mss;                                      \
   * mem_nb  = nb;                                       \
   */                                                    \
}

#define MEMORISE_TMP_IN_DISPLAY_AND_RESET()  {                                                                                  \
 if (mem_first_tuple_aligned != NULL) {                                                                                         \
   int left_score  = 0, left_pos_begin  = TBL_POS(mem_first_tuple_aligned), right_pos_begin = TBR_POS(mem_first_tuple_aligned); \
   int right_score = 0, left_pos_end    = TEL_POS(mem_last_tuple_aligned),  right_pos_end   = TER_POS(mem_last_tuple_aligned) ; \
                                                                                                                                \
   left_alignment_SG_Lz(                                                                                                        \
     data_query, TBL_POS(mem_first_tuple_aligned), MIN(TBL_POS(mem_first_tuple_aligned),gp_rho_stat),                           \
     data_text,  TBR_POS(mem_first_tuple_aligned), MIN(TBR_POS(mem_first_tuple_aligned),gp_rho_stat),                           \
     gp_xdrop, &left_score, &left_pos_begin, &right_pos_begin );                                                                \
   mem_total_score -= left_score;                                                                                               \
                                                                                                                                \
   right_alignment_SG_Lz(                                                                                                       \
     data_query, TEL_POS(mem_last_tuple_aligned), MIN(datasize_query - TEL_POS(mem_last_tuple_aligned) - 2, gp_rho_stat),       \
     data_text,  TER_POS(mem_last_tuple_aligned), MIN(datasize_text -  TER_POS(mem_last_tuple_aligned) - 2, gp_rho_stat),       \
     gp_xdrop, &right_score, &left_pos_end, &right_pos_end );                                                                   \
   mem_total_score -= right_score;                                                                                              \
                                                                                                                                \
   if (mem_total_score >  gp_minscore) {                                                                                        \
	 CreateMA(data_query, datasize_query,                                                                                   \
		  data_text,  datasize_text,                                                                                    \
                  left_pos_begin,                                                                                               \
		  right_pos_begin,                                                                                              \
		  left_pos_end,                                                                                                 \
		  right_pos_end,                                                                                                \
		  tuple_list,                                                                                                   \
		  mem_first_tuple_aligned,                                                                                      \
		  mem_last_tuple_aligned,                                                                                       \
		  mem_total_score,                                                                                              \
		  /*mem_mss, mem_Mss, mem_nb,*/                                                                                 \
		  gv_reverse_repeat,                                                                                            \
		  i_chunk);                                                                                                     \
    }                                                                                                                           \
    mem_first_tuple_aligned = NULL; 	                                                                                        \
 }                                                                                                                              \
 /* Reset all to 0  */                                                                                                          \
 mem_total_score = 0 ;                                                                                                          \
 total_score =  seedscore;                                                                                                      \
 if (t){							                                                                \
    first_tuple_aligned = t;                                                                                                    \
    last_tuple_aligned  = t;                                                                                                    \
    /* Mss = TSIZE(t);                                                                                                          \
     * mss = TSIZE(t);                                                                                                          \
     * nb = 1;                                                                                                                  \
     */                                                                                                                         \
 } else {                                                                                                                       \
    first_tuple_aligned = NULL;                                                                                                 \
    last_tuple_aligned  = NULL;                                                                                                 \
   /* Mss = 0;                                                                                                                  \
    * mss = 0;                                                                                                                  \
    * nb =  0;                                                                                                                  \
    */                                                                                                                          \
 }   								                                                                \
 IF_FLAGSAVE_MEMORISE_IN_TMP();                                                                                                 \
}





/***********************************************************/


#ifdef INLINE	
inline 
#endif	
int AlignTuples(tuplelist * tuple_list, 
		 char * data_query, int datasize_query,  
		 char * data_text,  int datasize_text,
		 int i_chunk)
{
  int left_gap,right_gap;
  int score = 0, seedscore = 0;
  int flagsave = 0, flagdrop = 0;
  
  
  /* first and last tuples in the currrent alignment */
  tuple * first_tuple_aligned =  tuple_list->first_tuple;
  tuple * last_tuple_aligned  =  tuple_list->first_tuple;
  tuple * t                   =  tuple_list->first_tuple->next;
   
  int total_score    = - alignment_SG_Strait(data_query, TBL_POS(tuple_list->first_tuple), 
					     data_text , TBR_POS(tuple_list->first_tuple), 
					     TSIZE(tuple_list->first_tuple));

  /*
   * int mss = TSIZE(tuple_list->first_tuple);
   * int Mss = TSIZE(tuple_list->first_tuple);
   * int nb = 1;
   */

  /* temporary mem vars */ 
  tuple * mem_first_tuple_aligned = NULL;
  tuple * mem_last_tuple_aligned  = NULL;
  int mem_total_score = 0;
  
  /*
   * int mem_mss         = 0;
   * int mem_Mss         = 0;
   * int mem_nb          = 0;
   */
  
  MEMORISE_IN_TMP();
  
  /* one chain alignment */
  
  while(t)
    {
    ta:

      right_gap  =  TGAP_L( last_tuple_aligned, t);
      left_gap   =  TGAP_R( last_tuple_aligned, t);

      if ( left_gap > 0 &&  right_gap  > 0 ){
	
	/* compute sub alignments scores */
	
	seedscore =  - alignment_SG_Strait(data_query, TBL_POS(t),
					   data_text , TBR_POS(t), 
					   TSIZE(t));

       
	
	if (left_gap == right_gap)
	  score =  - alignment_SG_Strait(data_query, TEL_POS(last_tuple_aligned),
					 data_text , TER_POS(last_tuple_aligned), 
					 left_gap);
	else 
	
	  score =  /* - alignment_SG_DROP_opt(data_query, TEL_POS(last_tuple_aligned),  left_gap,
					   data_text , TER_POS(last_tuple_aligned), right_gap,
					   seedscore + gp_xdrop);*/

	-alignment_SG_Border(
			     data_query, TEL_POS(last_tuple_aligned),  left_gap,
			     data_text , TER_POS(last_tuple_aligned), right_gap,
			     seedscore + gp_xdrop);
	
	total_score += score + seedscore;
	last_tuple_aligned  = t;
	/* 
	 *  Mss = MAX( Mss , TSIZE(t));
	 *  mss = MIN( mss , TSIZE(t));
	 *  nb++;
	 */ 

	
	
	IF_FLAGSAVE_MEMORISE_IN_TMP();
	IF_FLAGDROP_MEMORISE_TMP_IN_DISPLAY_AND_RESET();
	
	t = t ->next; 
	
      }else{
	/* some gaps are negatives ?? lead to alignments which are not coherents ...  */
	/* (1) next tuple does not extend any alignment ... removed ... */ 
	if (TEL_POS(t) < TEL_POS(last_tuple_aligned) || TER_POS(t) < TER_POS(last_tuple_aligned)){
	  last_tuple_aligned->next = t->next;
	  FreeTuple(t);
	  t = last_tuple_aligned->next;
	}else{
	  /* (2.1) next tuple increse the next alignment on both sequences but overlaps with previous part found */
	  if ((t->size) > 0){
	    (t->size)--;
	    goto ta;
	  }
	/* (2.2) problem with the overlap which cannot be solved in an efficient way so we remove this tuple ... */
	  last_tuple_aligned->next = t->next; 
	  FreeTuple(t);
	  t = last_tuple_aligned->next;
	}
      }/*if ( left_gap >= 0 &&  right_gap >= 0)       */
    }/*while(t)	 */
  /*  memorise possibly tmp into displayed elements */
  MEMORISE_TMP_IN_DISPLAY_AND_RESET();
  return 0;
}


int AlignAndFree(tuplelist * tuple_list, 
		 char * data_query, int datasize_query, 
		 char * data_text,  int datasize_text,
		 int force,int i_chunk){
  
  /* list of lists manipulating vars */
  tuplelist * tl_ = NULL , *tl_prev_ = NULL;
  tuple     * t_  = NULL;
  int       diag  = 0;
  int       countreturn = 8;
  /* statistical vars */
  int number_of_tuples = 0;
  int r_bitcover = 0;


  /* head lists selection and processing (alignement) :
   *
   *  Those lists are the older ones, so if i_liste < gv_i_current - threshold, 
   *  those lists will not be used by assemble
   *  we can process those and free some memory space
   */
   

  /*[a] tuplelists start  */
  tl_prev_ = tl_ = tuple_list;

  while (tl_ != NULL){
        
    t_ = tl_->first_tuple;
    r_bitcover = 0;
    number_of_tuples = 0;
    diag = 0;
    
    if (t_)
      diag = t_->distance;
    
    /* (1) list used by ASSEMBLE ? */
    while (t_ != NULL){
       if(t_->distance == diag){ 
	 r_bitcover += (t_->bitweight);
       }else{
	 r_bitcover = MAX(r_bitcover,t_->bitweight);
	 diag = t_->distance;
       }
      
      /* if list goes further, we do not consider it
       * and we stop this algorithm :
       * - all the following list have also chances to
       *   have their last element inside consideration
       *   zone
       */
      
      if (!force)
	if ( t_->occurrence > gv_i_current - gp_rho_stat - gp_delta_stat - gp_seed_span){
	  countreturn--;
	  if (!countreturn)
	    return 0;
	  else
	    goto nexttuplelist;
	}
      
      number_of_tuples++;
      t_ = t_->next;
    }/* (1) */

    /* (2) specific consideration of each tuple list */

    /* lists iteration has not given any element in the active update  
     * zone. Those alignements have been "completed" for the assemble 
     * algorithm and thus can be aligned
     */

    if (number_of_tuples <= 0){
      
      /* (2.0) The first list is always empty (do nothing 
       * and jump to the next list)
       */
    nexttuplelist:
      tl_prev_ = tl_ ;
      tl_      = tl_->next;
    }else{
      
      
     
      
      /* (2.1) alignements of one of several tuples
       */
        
      if (r_bitcover >= 2*gp_minimum_score_seeds){
#ifdef	STATS
	nb_chains_tested++;
#endif		
		
	AlignTuples(tl_,
		    data_query, datasize_query,
		    data_text , datasize_text,
		    i_chunk);
	
      }
      FreeTupleList(&tl_,&tl_prev_); 
    }
  }/* [a] tuple list iteration */

  return 0;
}

==> yass-1.08.tar.gz: yass-1.08/align.c <==
/*
 *  YASS 1.08
 *  Copyright (C) 2004,2005
 *  YASS Team
 *  Laurent Noe, Gregory Kucherov, Mikhail Roytberg, Christophe Valmir, Steven Corroy
 *  yass@loria.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This software contains code derived from the GNU libavl library.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <stdio.h>
#include <stdlib.h>

#include "align.h"
#include "kword.h"
#include "tuple.h"
#include "prdyn.h"
#include "display.h"
#include "util.h"
#include "global_var.h"


#define FLAGSAVE() {                                \
 flagsave = (total_score >= MAX(0,mem_total_score));\
                                                    \
}

#define FLAGDROP() {                                        \
  flagdrop = (total_score  + gp_xdrop <  mem_total_score ); \
                                                            \
                                                            \
}


#define IF_FLAGDROP_MEMORISE_TMP_IN_DISPLAY_AND_RESET(){    \
 FLAGDROP();                                                \
 if ( flagdrop ){                                           \
    MEMORISE_TMP_IN_DISPLAY_AND_RESET();                    \
 }                                                          \
}



#define IF_FLAGSAVE_MEMORISE_IN_TMP()  \
{                                      \
  FLAGSAVE();                          \
  if ( flagsave ) {                    \
     MEMORISE_IN_TMP();                \
  }                                    \
}


#define MEMORISE_IN_TMP() {                              \
  mem_first_tuple_aligned = first_tuple_aligned;         \
  mem_last_tuple_aligned  = last_tuple_aligned;          \
  mem_total_score         = total_score;                 \
  /* mem_Mss = Mss;                                      \
   * mem_mss = mss;                                      \
   * mem_nb  = nb;                                       \
   */                                                    \
}

#define MEMORISE_TMP_IN_DISPLAY_AND_RESET()  {                                                                                  \
 if (mem_first_tuple_aligned != NULL) {                                                                                         \
   int left_score  = 0, left_pos_begin  = TBL_POS(mem_first_tuple_aligned), right_pos_begin = TBR_POS(mem_first_tuple_aligned); \
   int right_score = 0, left_pos_end    = TEL_POS(mem_last_tuple_aligned),  right_pos_end   = TER_POS(mem_last_tuple_aligned) ; \
                                                                                                                                \
   left_alignment_SG_Lz(                                                                                                        \
     data_query, TBL_POS(mem_first_tuple_aligned), MIN(TBL_POS(mem_first_tuple_aligned),gp_rho_stat),                           \
     data_text,  TBR_POS(mem_first_tuple_aligned), MIN(TBR_POS(mem_first_tuple_aligned),gp_rho_stat),                           \
     gp_xdrop, &left_score, &left_pos_begin, &right_pos_begin );                                                                \
   mem_total_score -= left_score;                                                                                               \
                                                                                                                                \
   right_alignment_SG_Lz(                                                                                                       \
     data_query, TEL_POS(mem_last_tuple_aligned), MIN(datasize_query - TEL_POS(mem_last_tuple_aligned) - 2, gp_rho_stat),       \
     data_text,  TER_POS(mem_last_tuple_aligned), MIN(datasize_text -  TER_POS(mem_last_tuple_aligned) - 2, gp_rho_stat),       \
     gp_xdrop, &right_score, &left_pos_end, &right_pos_end );                                                                   \
   mem_total_score -= right_score;                                                                                              \
                                                                                                                                \
   if (mem_total_score >  gp_minscore) {                                                                                        \
	 CreateMA(data_query, datasize_query,                                                                                   \
		  data_text,  datasize_text,                                                                                    \
                  left_pos_begin,                                                                                               \
		  right_pos_begin,                                                                                              \
		  left_pos_end,                                                                                                 \
		  right_pos_end,                                                                                                \
		  tuple_list,                                                                                                   \
		  mem_first_tuple_aligned,                                                                                      \
		  mem_last_tuple_aligned,                                                                                       \
		  mem_total_score,                                                                                              \
		  /*mem_mss, mem_Mss, mem_nb,*/                                                                                 \
		  gv_reverse_repeat,                                                                                            \
		  i_chunk);                                                                                                     \
    }                                                                                                                           \
    mem_first_tuple_aligned = NULL; 	                                                                                        \
 }                                                                                                                              \
 /* Reset all to 0  */                                                                                                          \
 mem_total_score = 0 ;                                                                                                          \
 total_score =  seedscore;                                                                                                      \
 if (t){							                                                                \
    first_tuple_aligned = t;                                                                                                    \
    last_tuple_aligned  = t;                                                                                                    \
    /* Mss = TSIZE(t);                                                                                                          \
     * mss = TSIZE(t);                                                                                                          \
     * nb = 1;                                                                                                                  \
     */                                                                                                                         \
 } else {                                                                                                                       \
    first_tuple_aligned = NULL;                                                                                                 \
    last_tuple_aligned  = NULL;                                                                                                 \
   /* Mss = 0;                                                                                                                  \
    * mss = 0;                                                                                                                  \
    * nb =  0;                                                                                                                  \
    */                                                                                                                          \
 }   								                                                                \
 IF_FLAGSAVE_MEMORISE_IN_TMP();                                                                                                 \
}



/***********************************************************/


#ifdef INLINE	
inline 
#endif	
int AlignTuples(tuplelist * tuple_list, 
		 char * data_query, int datasize_query,  
		 char * data_text,  int datasize_text,
		 int i_chunk)
{
  int left_gap,right_gap;
  int score = 0, seedscore = 0;
  int flagsave = 0, flagdrop = 0;
  
  
  /* first and last tuples in the currrent alignment */
  tuple * first_tuple_aligned =  tuple_list->first_tuple;
  tuple * last_tuple_aligned  =  tuple_list->first_tuple;
  tuple * t                   =  tuple_list->first_tuple->next;
  
  int total_score = - alignment_SG_Strait(data_query, TBL_POS(tuple_list->first_tuple), 
					     data_text , TBR_POS(tuple_list->first_tuple), 
					     TSIZE(tuple_list->first_tuple));
  /*
   * int mss = TSIZE(tuple_list->first_tuple);
   * int Mss = TSIZE(tuple_list->first_tuple);
   * int nb = 1;
   */

  /* temporary mem vars */ 
  tuple * mem_first_tuple_aligned = NULL;
  tuple * mem_last_tuple_aligned  = NULL;
  int mem_total_score = 0;
  
  /*
   * int mem_mss         = 0;
   * int mem_Mss         = 0;
   * int mem_nb          = 0;
   */
  
  MEMORISE_IN_TMP();
  
  /* one chain alignment */
  
  while(t)
    {
    ta:

      right_gap  =  TGAP_L( last_tuple_aligned, t);
      left_gap   =  TGAP_R( last_tuple_aligned, t);

      if ( left_gap > 0 &&  right_gap  > 0 ){
	
	/* compute sub alignments scores */
	
	seedscore =  - alignment_SG_Strait(data_query, TBL_POS(t),
					   data_text , TBR_POS(t), 
					   TSIZE(t));
	
	if (left_gap == right_gap)
	  score =  - alignment_SG_Strait(data_query, TEL_POS(last_tuple_aligned),
					 data_text , TER_POS(last_tuple_aligned), 
					 left_gap);
	else 
	
	  score =  /* - alignment_SG_DROP_opt(data_query, TEL_POS(last_tuple_aligned),  left_gap,
					   data_text , TER_POS(last_tuple_aligned), right_gap,
					   seedscore + gp_xdrop);*/

	-alignment_SG_Border(
			     data_query, TEL_POS(last_tuple_aligned),  left_gap,
			     data_text , TER_POS(last_tuple_aligned), right_gap,
			     seedscore + gp_xdrop);
	
	total_score += score + seedscore;
	last_tuple_aligned  = t;
	/* 
	 *  Mss = MAX( Mss , TSIZE(t));
	 *  mss = MIN( mss , TSIZE(t));
	 *  nb++;
	 */ 

	
	
	IF_FLAGSAVE_MEMORISE_IN_TMP();
	IF_FLAGDROP_MEMORISE_TMP_IN_DISPLAY_AND_RESET();
	
	t = t ->next; 
	
      }else{
	/* some gaps are negatives ?? lead to alignments which are not coherents ...  */
	/* (1) next tuple does not extend any alignment ... removed ... */
	if (TEL_POS(t) < TEL_POS(last_tuple_aligned) || TER_POS(t) < TER_POS(last_tuple_aligned)){
	  last_tuple_aligned->next = t->next;
	  FreeTuple(t);
	  t = last_tuple_aligned->next;

	}else{
	  /* (2.1) next tuple increse the next alignment on both sequences but overlaps with previous part found */
	  if ((t->size) > 0){
	    (t->size)--;
	    if (t->size<0) fprintf(stderr,"ERREUR Tuple negatif dans assemble\n");
	    goto ta;
	  }
	/* (2.2) problem with the overlap which cannot be solved in an efficient way so we remove this tuple ... */
	  last_tuple_aligned->next = t->next; 
	  FreeTuple(t);
	  t = last_tuple_aligned->next;
	}
      }/*if ( left_gap >= 0 &&  right_gap >= 0)       */
    }/*while(t)	 */
  /*  memorise possibly tmp into displayed elements */
  MEMORISE_TMP_IN_DISPLAY_AND_RESET();
  return 0;
}

int AlignAndFree(tuplelist * tuple_list, 
		 char * data_query, int datasize_query, 
		 char * data_text,  int datasize_text,
		 int force,int i_chunk) {
  
  /* list of lists manipulating vars */
  tuplelist * tl_ = NULL , *tl_prev_ = NULL;
  tuple     * t_  = NULL;
  int       diag  = 0;
  int       countreturn = 8;
  /* statistical vars */
  int number_of_tuples = 0;
  int r_bitcover = 0;


  /* head lists selection and processing (alignement) :
   *
   *  Those lists are the older ones, so if i_liste < gv_i_current - threshold, 
   *  those lists will not be used by assemble
   *  we can process those and free some memory space
   */
   

  /*[a] tuplelists start  */
  tl_prev_ = tl_ = tuple_list;

  while (tl_ != NULL){
    
    
    t_ = tl_->first_tuple;
    r_bitcover = 0;
    number_of_tuples = 0;
    diag = 0;
    
    if (t_)
      diag = t_->distance;
    
    /* (1) list used by ASSEMBLE ? */
    while (t_ != NULL){
       if(t_->distance == diag){ 
	 r_bitcover += (t_->bitweight);
       }else{
	 r_bitcover = MAX(r_bitcover,t_->bitweight);
	 diag = t_->distance;
       }
      
      /* if list goes further, we do not consider it
       * and we stop this algorithm :
       * - all the following list have also chances to
       *   have their last element inside consideration
       *   zone
       */
      
      if (!force)
	if ( t_->occurrence > gv_i_current - gp_rho_stat - gp_delta_stat - gp_seeds_span_min){
	  countreturn--;
	  if (!countreturn)
	    return 0;
	  else
	    goto nexttuplelist;
	}
      
      number_of_tuples++;
      t_ = t_->next;
    }/* (1) */

    /* (2) specific consideration of each tuple list */

    /* lists iteration has not given any element in the active update  
     * zone. Those alignements have been "completed" for the assemble 
     * algorithm and thus can be aligned
     */

    if (number_of_tuples <= 0){
      
      /* (2.0) The first list is always empty (do nothing 
       * and jump to the next list)
       */
    nexttuplelist:
      tl_prev_ = tl_ ;
      tl_      = tl_->next;
    }else{
      
      
     
      
      /* (2.1) alignements of one of several tuples
       */
        
      if (r_bitcover >= 2*gp_minimum_score_seeds){
#ifdef	STATS
	nb_chains_tested++;
#endif		
		
	AlignTuples(tl_,
		    data_query, datasize_query,
		    data_text , datasize_text,
		    i_chunk);	
      }
      FreeTupleList(&tl_,&tl_prev_); 
    }
  }/* [a] tuple list iteration */

  return 0;
}
